#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


import unittest
from utils import initialise_tests

import numpy as np

from cil.framework import VectorData


from cil.optimisation.functions import LeastSquares
from cil.optimisation.functions import ApproximateGradientSumFunction
from cil.optimisation.functions import SGFunction, SAGFunction, SAGAFunction
from cil.optimisation.functions import SVRGFunction, LSVRGFunction
from cil.optimisation.operators import MatrixOperator
from cil.optimisation.algorithms import GD
from cil.framework import VectorData
from cil.optimisation.utilities import Sampler, SamplerRandom

from testclass import CCPiTestClass
from utils import has_astra

initialise_tests()

if has_astra:
    from cil.plugins.astra import ProjectionOperator

from utils import has_cvxpy

if has_cvxpy:
    import cvxpy


class TestApproximateGradientSumFunction(CCPiTestClass):

    def setUp(self):
        self.sampler = Sampler.random_with_replacement(5)
        self.initial = VectorData(np.zeros(10))
        self.b = VectorData(np.random.normal(0, 1, 10))
        self.functions = []
        for i in range(5):
            diagonal = np.zeros(10)
            diagonal[2*i:2*(i+1)] = 1
            A = MatrixOperator(np.diag(diagonal))
            self.functions.append(LeastSquares(A, A.direct(self.b)))
            if i == 0:
                self.objective = LeastSquares(A, A.direct(self.b))
            else:
                self.objective += LeastSquares(A, A.direct(self.b))

    def test_ABC(self):
        with self.assertRaises(TypeError):
            self.stochastic_objective = ApproximateGradientSumFunction(
                self.functions, self.sampler)



class approx_gradient_child_class_testing():

    def set_up(self):
        np.random.seed(10)
        self.sampler = Sampler.random_with_replacement(self.n_subsets)
        self.initial = VectorData(np.zeros(30))
        b = VectorData(np.array(range(30))/50)

        self.f_subsets = []
        for i in range(self.n_subsets):
            diagonal = np.zeros(30)
            diagonal[(30//self.n_subsets)*i:(30//self.n_subsets)*(i+1)] = 1
            Ai = MatrixOperator(np.diag(diagonal))
            self.f_subsets.append(LeastSquares(Ai, Ai.direct(b)))
        self.A=MatrixOperator(np.diag(np.ones(30)))
        self.f = LeastSquares(self.A, b)

        self.f_stochastic=self.stochastic_estimator(self.f_subsets, self.sampler)

    def test_approximate_gradient_not_equal_full(self):
        self.f_stochastic.gradient(self.initial)
        self.assertFalse((self.f_stochastic.full_gradient(
            self.initial+1) == self.f_stochastic.gradient(self.initial+1).array).all())

    def test_sampler(self):
        self.assertTrue(isinstance(self.f_stochastic.sampler, SamplerRandom))
        f = self.stochastic_estimator(self.f_subsets)
        self.assertTrue(isinstance(f.sampler, SamplerRandom))
        self.assertEqual(f.sampler._type, 'random_with_replacement')

    def test_call(self):
        self.assertAlmostEqual(self.f_stochastic(
            self.initial), self.f(self.initial), 1)

    def test_full_gradient(self):
        self.assertNumpyArrayAlmostEqual(self.f_stochastic.full_gradient(
            self.initial).array, self.f.gradient(self.initial).array, 2)

    def test_error_without_function(self):
        self.stochastic_estimator([self.f], self.sampler)
        with self.assertRaises((IndexError, ZeroDivisionError)):
            self.stochastic_estimator([], self.sampler)

    def test_type_error_if_functions_not_a_list(self):
        with self.assertRaises(TypeError):
            self.stochastic_estimator(self.f, self.sampler)

    def test_sampler_without_next(self):
        class bad_Sampler():
            def init(self):
                pass
        bad_sampler = bad_Sampler()
        with self.assertRaises(ValueError):
            self.stochastic_estimator([self.f, self.f], bad_sampler)


    def test_sampler_out_of_range(self):
        def g(index):
            return -2
        bad_sampler = Sampler.from_function(12,g)
        f = self.stochastic_estimator([self.f]*10, bad_sampler)
        with self.assertRaises(IndexError):
            f.gradient(self.initial)
            f.gradient(self.initial)

    @unittest.skipUnless(has_cvxpy, "CVXpy not installed")
    def test_toy_example(self):
        sampler = Sampler.random_with_replacement(5)
        initial = VectorData(np.zeros(25))
        b = VectorData(np.array(range(25)))
        functions = []
        for i in range(5):
            diagonal = np.zeros(25)
            diagonal[5*i:5*(i+1)] = 1
            A = MatrixOperator(np.diag(diagonal))
            functions.append(0.5*LeastSquares(A, A.direct(b)))

        Aop=MatrixOperator(np.diag(np.ones(25)))

        u_cvxpy = cvxpy.Variable(b.shape[0])
        objective = cvxpy.Minimize( 0.5*cvxpy.sum_squares(Aop.A @ u_cvxpy - Aop.direct(b).array))
        p = cvxpy.Problem(objective)
        p.solve(verbose=True, solver=cvxpy.SCS, eps=1e-4)

        stochastic_objective = self.stochastic_estimator(functions, sampler)

        alg_stochastic = GD(initial=initial,
                            f=stochastic_objective, update_objective_interval=1000,
                            step_size=0.05)
        alg_stochastic.run(600, verbose=0)

        np.testing.assert_allclose(p.value ,stochastic_objective(alg_stochastic.x) , atol=1e-1)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), u_cvxpy.value, 3)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), b.as_array(), 3)


class TestSGD(CCPiTestClass, approx_gradient_child_class_testing):

    def setUp(self):
        self.stochastic_estimator=SGFunction
        self.n_subsets=6
        self.set_up()

    def test_partition_weights(self):
        f_stochastic=SGFunction(self.f_subsets, Sampler.sequential(self.n_subsets))
        self.assertListEqual(f_stochastic._partition_weights, [1 / self.n_subsets] * self.n_subsets)
        with self.assertRaises(ValueError):
            f_stochastic.set_data_partition_weights( list(range(self.n_subsets)))
        with self.assertRaises(ValueError):
            f_stochastic.set_data_partition_weights( [1])
        with self.assertRaises(ValueError):
            f_stochastic.set_data_partition_weights( [-1]+[2/(self.n_subsets-1)]*(self.n_subsets-1))
        a=[i/float(sum(range(self.n_subsets))) for i in range(self.n_subsets)]
        f_stochastic.set_data_partition_weights( a)
        self.assertListEqual(f_stochastic._partition_weights, a )
        f_stochastic.gradient(self.initial)
        for i in range(1,20):
            f_stochastic.gradient(self.initial)
            self.assertEqual(f_stochastic.data_passes[i], f_stochastic.data_passes[i-1]+a[i%self.n_subsets])




class TestSAG(CCPiTestClass, approx_gradient_child_class_testing):

    def setUp(self):
        self.stochastic_estimator=SAGFunction
        self.n_subsets=6
        self.set_up()

    def test_warm_start_and_data_passes(self):
        f1=SAGFunction(self.f_subsets,Sampler.sequential(self.n_subsets))
        f=SAGFunction(self.f_subsets,Sampler.sequential(self.n_subsets))
        f.warm_start_approximate_gradients(self.initial)
        f1.gradient(self.initial)
        f.gradient(self.initial)
        self.assertEqual(f.function_num, 0)
        self.assertEqual(f1.function_num, 0)
        self.assertNumpyArrayAlmostEqual(np.array(f1.data_passes), np.array([1./f1.num_functions]))
        self.assertNumpyArrayAlmostEqual(np.array(f.data_passes), np.array([ 1+1./f1.num_functions]))
        self.assertNumpyArrayAlmostEqual(np.array(f.data_passes_indices[0]), np.array(list(range(f1.num_functions))+ [0]))
        self.assertNumpyArrayAlmostEqual(np.array(f1.data_passes_indices[0]), np.array([0]))
        self.assertNumpyArrayAlmostEqual(f._list_stored_gradients[0].array, f1._list_stored_gradients[0].array)
        self.assertNumpyArrayAlmostEqual(f._list_stored_gradients[0].array, self.f_subsets[0].gradient(self.initial).array)
        self.assertNumpyArrayAlmostEqual(f._list_stored_gradients[1].array, self.f_subsets[1].gradient(self.initial).array)

        self.assertFalse((f._list_stored_gradients[3].array== f1._list_stored_gradients[3].array).all())

    @unittest.skipUnless(has_cvxpy, "CVXpy not installed")
    def test_SAG_toy_example_warm_start(self):
        sampler=Sampler.random_with_replacement(3,seed=1)
        initial = VectorData(np.zeros(21))
        np.random.seed(4)
        b =  VectorData(np.random.normal(0,4,21))
        functions=[]
        for i in range(3):
            diagonal=np.zeros(21)
            diagonal[7*i:7*(i+1)]=1
            A=MatrixOperator(np.diag(diagonal))
            functions.append( LeastSquares(A, A.direct(b)))

        Aop=MatrixOperator(np.diag(np.ones(21)))

        u_cvxpy = cvxpy.Variable(b.shape[0])
        objective = cvxpy.Minimize( 0.5*cvxpy.sum_squares(Aop.A @ u_cvxpy - Aop.direct(b).array))
        p = cvxpy.Problem(objective)
        p.solve(verbose=True, solver=cvxpy.SCS, eps=1e-4)

        stochastic_objective=SAGFunction(functions, sampler)
        stochastic_objective.warm_start_approximate_gradients(initial)

        alg_stochastic = GD(initial=initial,
                              f=stochastic_objective, update_objective_interval=1000,
                              step_size=0.05, max_iteration =5000)
        alg_stochastic.run( 80, verbose=0)
        np.testing.assert_allclose(p.value ,stochastic_objective(alg_stochastic.x) , atol=1e-1)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), u_cvxpy.value, 3)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), b.as_array(), 3)


class TestSAGA(CCPiTestClass,approx_gradient_child_class_testing):

    def setUp(self):
        self.stochastic_estimator=SAGAFunction
        self.n_subsets=6
        self.set_up()

    def test_warm_start_and_data_passes(self):
        f1=SAGAFunction(self.f_subsets,Sampler.sequential(self.n_subsets))
        f=SAGAFunction(self.f_subsets,Sampler.sequential(self.n_subsets))
        f.warm_start_approximate_gradients(self.initial)
        f1.gradient(self.initial)
        f.gradient(self.initial)

        self.assertEqual(f.function_num, 0)
        self.assertEqual(f1.function_num, 0)
        self.assertNumpyArrayAlmostEqual(np.array(f1.data_passes), np.array([1./f1.num_functions]))
        self.assertNumpyArrayAlmostEqual(np.array(f.data_passes), np.array([1+1./f1.num_functions]))
        self.assertNumpyArrayAlmostEqual(np.array(f.data_passes_indices[0]),np.array( list(range(self.n_subsets))+[0]))
        self.assertNumpyArrayAlmostEqual(np.array(f1.data_passes_indices[0]), np.array([0]))
        self.assertNumpyArrayAlmostEqual(f._list_stored_gradients[0].array, f1._list_stored_gradients[0].array)
        self.assertNumpyArrayAlmostEqual(f._list_stored_gradients[0].array, self.f_subsets[0].gradient(self.initial).array)
        self.assertNumpyArrayAlmostEqual(f._list_stored_gradients[1].array, self.f_subsets[1].gradient(self.initial).array)

        self.assertFalse((f._list_stored_gradients[1].array== f1._list_stored_gradients[1].array).all())
        self.assertNumpyArrayAlmostEqual(f1._list_stored_gradients[1].array, self.initial.array)

    @unittest.skipUnless(has_cvxpy, "CVXpy not installed")
    def test_SAGA_toy_example_warm_start(self):
        sampler=Sampler.random_with_replacement(3,seed=1)
        initial = VectorData(np.zeros(21))
        np.random.seed(4)
        b =  VectorData(np.random.normal(0,4,21))
        functions=[]
        for i in range(3):
            diagonal=np.zeros(21)
            diagonal[7*i:7*(i+1)]=1
            A=MatrixOperator(np.diag(diagonal))
            functions.append( LeastSquares(A, A.direct(b)))

        Aop=MatrixOperator(np.diag(np.ones(21)))

        u_cvxpy = cvxpy.Variable(b.shape[0])
        objective = cvxpy.Minimize( 0.5*cvxpy.sum_squares(Aop.A @ u_cvxpy - Aop.direct(b).array))
        p = cvxpy.Problem(objective)
        p.solve(verbose=True, solver=cvxpy.SCS, eps=1e-4)

        stochastic_objective=SAGAFunction(functions, sampler)
        stochastic_objective.warm_start_approximate_gradients(initial)

        alg_stochastic = GD(initial=initial,
                              f=stochastic_objective, update_objective_interval=1000,
                              step_size=0.05, max_iteration =5000)
        alg_stochastic.run( 100, verbose=0)
        np.testing.assert_allclose(p.value ,stochastic_objective(alg_stochastic.x) , atol=1e-1)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), u_cvxpy.value, 3)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), b.as_array(), 3)
        

class TestSVRG(CCPiTestClass, approx_gradient_child_class_testing):

    def setUp(self):
        self.stochastic_estimator=SVRGFunction
        self.n_subsets=5
        self.set_up()

    

    def test_SVRG_init(self):
        self.assertEqual(self.f_stochastic.snapshot_update_interval,
                         2*self.f_stochastic.num_functions)
        self.assertListEqual(self.f_stochastic.data_passes, [])
        self.assertListEqual(self.f_stochastic.data_passes_indices, [])
        self.assertEqual(self.f_stochastic.store_gradients, False)
        f2 = SVRGFunction(self.f_subsets, snapshot_update_interval=2,
                          store_gradients=True)
        self.assertEqual(f2.snapshot_update_interval, 2)
        self.assertListEqual(f2.data_passes, [])
        self.assertListEqual(f2.data_passes_indices, [])
        self.assertEqual(f2.store_gradients, True)

    def test_SVRG_snapshot_update_interval_and_data_passes(self):
        sampler = Sampler.random_with_replacement(self.n_subsets)
        objective = SVRGFunction(self.f_subsets, sampler)
        alg_stochastic = GD(initial=self.initial,
                            f=objective, update_objective_interval=500,
                            step_size=5e-8, max_iteration=5000)
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(
            np.array(objective.data_passes), np.array([1., 6./5]))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(
            np.array(objective.data_passes), np.array([1., 6./5, 7./5, 8./5]))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes), np.array([
                                         1., 6./5, 7./5, 8./5, 9./5, 10./5]))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes), np.array(
            [1., 6./5, 7./5, 8./5, 9./5, 10./5, 11./5, 12./5]))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes), np.array(
            [1., 6./5, 7./5, 8./5, 9./5, 10./5, 11./5, 12./5, 13./5, 14./5]))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes), np.array(
            [1., 6./5, 7./5, 8./5, 9./5, 10./5, 11./5, 12./5, 13./5, 14./5, 19./5, 20./5]))
        objective = SVRGFunction(
            self.f_subsets, self.sampler, snapshot_update_interval=3)
        alg_stochastic = GD(initial=self.initial,
                            f=objective, update_objective_interval=500,
                            step_size=5e-8, max_iteration=5000)
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(
            np.array(objective.data_passes), np.array([1., 6./5]))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(
            np.array(objective.data_passes), np.array([1., 6./5, 7./5, 12./5]))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes), np.array([
                                         1., 6./5, 7./5, 12./5, 13./5, 14./5]))
        alg_stochastic.run(2, verbose=0)

    def test_SVRG_store_gradients(self):
        objective = SVRGFunction(self.f_subsets, self.sampler)
        self.assertEqual(objective._list_stored_gradients, None)
        objective.gradient(self.initial)
        self.assertEqual(objective._list_stored_gradients, None)

        objective = SVRGFunction(
            self.f_subsets, self.sampler, store_gradients=True)
        self.assertEqual(objective._list_stored_gradients, None)
        objective.gradient(self.initial)
        self.assertNumpyArrayAlmostEqual(
            objective._list_stored_gradients[0].array, self.f_subsets[0].gradient(self.initial).array)
        self.assertNumpyArrayAlmostEqual(
            objective._list_stored_gradients[1].array, self.f_subsets[1].gradient(self.initial).array)


    @unittest.skipUnless(has_cvxpy, "CVXpy not installed") 
    def test_SVRG_toy_example_store_gradients(self):
        sampler = Sampler.sequential(3)
        initial = VectorData(np.zeros(15))
        np.random.seed(4)
        b = VectorData(np.random.normal(0, 3, 15))
        functions = []
        for i in range(3):
            diagonal = np.zeros(15)
            diagonal[5*i:5*(i+1)] = 1
            A = MatrixOperator(np.diag(diagonal))
            functions.append(LeastSquares(A, A.direct(b)))
        Aop=MatrixOperator(np.diag(np.ones(15)))

        u_cvxpy = cvxpy.Variable(b.shape[0])
        objective = cvxpy.Minimize( 0.5*cvxpy.sum_squares(Aop.A @ u_cvxpy - Aop.direct(b).array))
        p = cvxpy.Problem(objective)
        p.solve(verbose=True, solver=cvxpy.SCS, eps=1e-4) 
        
        
        stochastic_objective = SVRGFunction(functions, sampler, store_gradients=True)

        alg_stochastic = GD(initial=initial,
                            f=stochastic_objective, update_objective_interval=1000,
                            step_size=1/stochastic_objective.L)

        alg_stochastic.run(10, verbose=0)
        self.assertNumpyArrayAlmostEqual(np.array(stochastic_objective.data_passes), 
           np.array( [1.]+[4./3, 5./3, 6./3, 7./3,  8./3, 11./3, 12./3, 13./3, 14./3]), 4)
        self.assertListEqual(stochastic_objective.data_passes_indices[:2], [list(range(3)), [0]])

        alg_stochastic.run(100, verbose=0)
        np.testing.assert_allclose(p.value ,stochastic_objective(alg_stochastic.x) , atol=1e-1)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), u_cvxpy.value, 3)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), b.as_array(), 3)

    

class TestLSVRG(CCPiTestClass, approx_gradient_child_class_testing):

    def setUp(self):
        self.stochastic_estimator=LSVRGFunction
        self.n_subsets=5
        self.set_up()

    def test_approximate_gradient_not_equal_full(self):
        self.f_stochastic.gradient(self.initial)
        self.assertFalse((self.f_stochastic.full_gradient(
            self.initial+1) == self.f_stochastic.approximate_gradient(self.initial+1,3).array).all())


    def test_LSVRG_init(self):
        self.assertEqual(self.f_stochastic.snapshot_update_probability, 1/self.n_subsets)
        self.assertListEqual(self.f_stochastic.data_passes, [])
        self.assertListEqual(self.f_stochastic.data_passes_indices, [])
        self.assertEqual(self.f_stochastic.store_gradients, False)

        f2 = LSVRGFunction(self.f_subsets, snapshot_update_probability=1 /
                           2, store_gradients=True, seed=1)
        self.assertEqual(f2.snapshot_update_probability, 1/2)
        self.assertListEqual(f2.data_passes, [])
        self.assertEqual(f2.store_gradients, True)

    def test_LSVRG_data_passes_and_snapshot_update_probability_and_seed(self):
        objective = LSVRGFunction(self.f_subsets, self.sampler, snapshot_update_probability=1)
        alg_stochastic = GD(initial=self.initial,  update_objective_interval=500,
                            f=objective,                               step_size=5e-8, max_iteration=5000)
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(
            np.array(objective.data_passes), np.array([1., 2,]))
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes_indices[-1]), np.array(list(range(self.n_subsets))))
        alg_stochastic.run(2, verbose=0)
        self.assertNumpyArrayAlmostEqual(
            np.array(objective.data_passes), np.array([1., 2., 3., 4.]))
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes_indices[-1]), np.array(list(range(self.n_subsets))))
        objective = LSVRGFunction(self.f_subsets, self.sampler, seed=3)
        alg_stochastic = GD(initial=self.initial,
                            f=objective, update_objective_interval=500,
                            step_size=5e-8, max_iteration=5000)
        alg_stochastic.run(10, verbose=0)
        self.assertNumpyArrayAlmostEqual(np.array(objective.data_passes), 
            np.array([1., 2., 2.2, 2.4, 2.6, 3.6, 3.8, 4., 5., 5.2]))

    def test_LSVRG_store_gradients(self):
        objective = LSVRGFunction(self.f_subsets, self.sampler)
        
        self.assertEqual(objective._list_stored_gradients, None)
        objective.gradient(self.initial)
        self.assertEqual(objective._list_stored_gradients, None)

        objective = LSVRGFunction(
            self.f_subsets, self.sampler, store_gradients=True)
        self.assertEqual(objective._list_stored_gradients, None)
        objective.gradient(self.initial)
        self.assertNumpyArrayAlmostEqual(
            objective._list_stored_gradients[0].array, self.f_subsets[0].gradient(self.initial).array)
        self.assertNumpyArrayAlmostEqual(
            objective._list_stored_gradients[1].array, self.f_subsets[1].gradient(self.initial).array)




    @unittest.skipUnless(has_cvxpy, "CVXpy not installed") 
    def test_LSVRG_toy_example_store_gradients(self):
        sampler = Sampler.sequential(3)
        initial = VectorData(np.zeros(15))
        np.random.seed(4)
        b = VectorData(np.random.normal(0, 3, 15))
        functions = []
        for i in range(3):
            diagonal = np.zeros(15)
            diagonal[5*i:5*(i+1)] = 1
            A = MatrixOperator(np.diag(diagonal))
            functions.append(LeastSquares(A, A.direct(b)))
        Aop=MatrixOperator(np.diag(np.ones(15)))

        u_cvxpy = cvxpy.Variable(b.shape[0])
        objective = cvxpy.Minimize( 0.5*cvxpy.sum_squares(Aop.A @ u_cvxpy - Aop.direct(b).array))
        p = cvxpy.Problem(objective)
        p.solve(verbose=True, solver=cvxpy.SCS, eps=1e-4) 
        
        
        stochastic_objective = LSVRGFunction(functions, sampler, store_gradients=True)

        alg_stochastic = GD(initial=initial,
                            f=stochastic_objective, update_objective_interval=1000,
                            step_size=1/stochastic_objective.L)



        alg_stochastic.run(100, verbose=0)
        np.testing.assert_allclose(p.value ,stochastic_objective(alg_stochastic.x) , atol=1e-1)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), u_cvxpy.value, 3)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), b.as_array(), 3)
        
    def test_sampler_out_of_range(self):
        def g(index):
            return -2
        bad_sampler = Sampler.from_function(12,g)
        f = self.stochastic_estimator([self.f]*10, bad_sampler, seed=25)
        with self.assertRaises(IndexError):
            f.gradient(self.initial)
            f.gradient(self.initial)
            


#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from unittest.mock import Mock, patch
import logging
import numpy
from cil.framework import ImageGeometry
from cil.optimisation.operators import GradientOperator
from cil.optimisation.operators import LinearOperator
from utils import initialise_tests

log = logging.getLogger(__name__)
initialise_tests()


class TestGradientOperator(unittest.TestCase):

    def setUp(self, *args, **kwargs):
        N, M, K = 10, 11, 12
        channels = 13

        self.voxel_size_x, self.voxel_size_y, self.voxel_size_z = 0.1, 0.3, 0.6

        self.ig_2D = ImageGeometry(voxel_num_x = M, voxel_num_y = N)
        self.ig_2D_chan = ImageGeometry(voxel_num_x = M, voxel_num_y = N, channels = channels)
        self.ig_3D = ImageGeometry(voxel_num_x = M, voxel_num_y = N, voxel_num_z= K)
        self.ig_3D_chan = ImageGeometry(voxel_num_x = M, voxel_num_y = N, channels = channels, voxel_num_z= K)

        self.ig_2D_voxel = ImageGeometry(voxel_num_x = M, voxel_num_y = N, voxel_size_x=self.voxel_size_x, voxel_size_y=self.voxel_size_y)
        self.ig_2D_chan_voxel = ImageGeometry(voxel_num_x = M, voxel_num_y = N, channels = channels, voxel_size_x=self.voxel_size_x, voxel_size_y=self.voxel_size_y)
        self.ig_3D_voxel = ImageGeometry(voxel_num_x = M, voxel_num_y = N, voxel_num_z= K, voxel_size_x=self.voxel_size_x, voxel_size_y=self.voxel_size_y, voxel_size_z=self.voxel_size_z)
        self.ig_3D_chan_voxel = ImageGeometry(voxel_num_x = M, voxel_num_y = N, channels = channels, voxel_num_z= K, voxel_size_y=self.voxel_size_y, voxel_size_x=self.voxel_size_x, voxel_size_z=self.voxel_size_z)

        self.list_geometries = [self.ig_2D, self.ig_2D_chan, self.ig_3D, self.ig_3D_chan,
                                self.ig_2D_voxel, self.ig_2D_chan_voxel, self.ig_3D_voxel, self.ig_3D_chan_voxel]

        test_matrix_backend_numpy = {  'backend':'numpy',
                                            'bconditions': ['Neumann', 'Periodic'],
                                            'correlation':['Space','SpaceChannels'],
                                            'method':['forward', 'backward', 'centered']}

        test_matrix_backend_c ={   'backend':'c',
                                        'bconditions': ['Neumann', 'Periodic'],
                                        'correlation':['SpaceChannels'],
                                        'method':['forward']}

        self.test_backend_configurations = [test_matrix_backend_numpy, test_matrix_backend_c]

    def print_assertion_info(self, geom = None, bnd = None, backend = None, method = None, corr = None, split = None):

        log.info("Test Failed")
        log.info("ImageGeometry %s", geom)
        log.info("Bnd Cond %s", bnd)
        log.info("Backend %s", backend)
        log.info("Method %s", method)
        log.info("Correlation %s", corr)
        if split is not None:
            log.info("Split %s", split)


    def test_GradientOperator_linearity(self):

        for config in self.test_backend_configurations:
            backend = config.get('backend')

            for geom in self.list_geometries:
                for bnd in config.get('bconditions'):
                    for corr in config.get('correlation'):
                        for method in config.get('method'):

                            Grad = GradientOperator(geom,
                                                    bnd_cond = bnd,
                                                    backend = backend,
                                                    correlation = corr, method = method)
                            try:
                                for sd in [5, 10, 15]:
                                    self.assertTrue(LinearOperator.dot_test(Grad, seed=sd))
                            except:
                                self.print_assertion_info(geom,bnd,backend,corr,method,None)
                                raise


    def test_GradientOperator_norm(self):

        for config in self.test_backend_configurations:
            backend = config.get('backend')

            for geom in self.list_geometries:
                for bnd in config.get('bconditions'):
                    for corr in config.get('correlation'):
                        for method in config.get('method'):

                            if geom.channels == 1:

                                if geom.length == 2:
                                    norm = numpy.sqrt(
                                        (2/geom.voxel_size_y)**2 + (2/geom.voxel_size_x)**2)
                                elif geom.length == 3:
                                    norm = numpy.sqrt(
                                        (2/geom.voxel_size_z)**2 + (2/geom.voxel_size_y)**2 + (2/geom.voxel_size_x)**2)

                            else:

                                if corr == 'Space':
                                    if geom.length ==3:
                                        norm = numpy.sqrt(
                                            (2/geom.voxel_size_y)**2 + (2/geom.voxel_size_x)**2)
                                    else:
                                        norm = numpy.sqrt((2/geom.voxel_size_z)**2 + (2/geom.voxel_size_y)**2 + (2/geom.voxel_size_x)**2)
                                else:

                                    if geom.length ==3:
                                        norm = numpy.sqrt(
                                            (2/geom.channel_spacing)**2 + (2/geom.voxel_size_y)**2 + (2/geom.voxel_size_x)**2)
                                    else:
                                        norm = numpy.sqrt((2/geom.channel_spacing)**2 + (2/geom.voxel_size_z)**2 + (2/geom.voxel_size_y)**2 + (2/geom.voxel_size_x)**2)


                            Grad = GradientOperator(geom,
                                                    bnd_cond= bnd,
                                                    backend = backend,
                                                    correlation= corr, method=method)
                            try:
                                self.assertAlmostEqual(Grad.norm(), norm, 6)
                            except AssertionError:
                                self.print_assertion_info(geom, bnd,backend,corr,method,None)
                                raise


    def test_GradientOperator_in_place_vs_allocate_direct(self):

        for config in self.test_backend_configurations:
            backend = config.get('backend')

            for geom in self.list_geometries:
                for bnd in config.get('bconditions'):
                    for corr in config.get('correlation'):
                        for method in config.get('method'):

                            Grad = GradientOperator(geom,
                                                    bnd_cond = bnd,
                                                    backend = backend,
                                                    correlation = corr, method=method)
                            tmp_x = Grad.domain.allocate('random')
                            res_in_place = Grad.direct(tmp_x)
                            res_allocate = Grad.range.allocate()
                            Grad.direct(tmp_x, out = res_allocate)

                            for m in range(len(Grad.range.geometries)):

                                try:
                                    numpy.testing.assert_array_almost_equal(res_in_place[m].array,
                                                                    res_allocate[m].array)
                                except:
                                    self.print_assertion_info(geom,bnd,backend,corr,method,None)
                                    raise


    def test_GradientOperator_in_place_vs_allocate_adjoint(self):

        for config in self.test_backend_configurations:
            backend = config.get('backend')

            for geom in self.list_geometries:
                for bnd in config.get('bconditions'):
                    for corr in config.get('correlation'):
                        for method in config.get('method'):

                            Grad = GradientOperator(geom,
                                                    bnd_cond = bnd,
                                                    backend = backend,
                                                    correlation = corr, method=method)
                            tmp_x = Grad.range.allocate('random')
                            res_in_place = Grad.adjoint(tmp_x)
                            res_allocate = Grad.domain.allocate()
                            Grad.adjoint(tmp_x, out = res_allocate)

                            try:
                                numpy.testing.assert_array_almost_equal(res_in_place.array,
                                                                res_allocate.array)
                            except:
                                self.print_assertion_info(geom,bnd,backend,corr,method, None)
                                raise


    def test_GradientOperator_range_shape(self):
        Grad2D = GradientOperator(self.ig_2D, backend='numpy')
        numpy.testing.assert_equal(Grad2D.range.shape, (2,1))

        Grad2D = GradientOperator(self.ig_2D_chan, correlation="Space", backend='numpy')
        numpy.testing.assert_equal(Grad2D.range.shape, (2,1))

        Grad2D_chan = GradientOperator(self.ig_2D_chan, correlation="SpaceChannels", backend='numpy')
        numpy.testing.assert_equal(Grad2D_chan.range.shape, (3,1))

        Grad3D = GradientOperator(self.ig_3D, backend='numpy')
        numpy.testing.assert_equal(Grad3D.range.shape, (3,1))

        Grad3D_chan = GradientOperator(self.ig_3D_chan, correlation="Space", backend='numpy')
        numpy.testing.assert_equal(Grad3D_chan.range.shape, (3,1))

        Grad3D_chan = GradientOperator(self.ig_3D_chan, correlation="SpaceChannels", backend='numpy')
        numpy.testing.assert_equal(Grad3D_chan.range.shape, (4,1))

        Grad2D = GradientOperator(self.ig_2D, backend='c')
        numpy.testing.assert_equal(Grad2D.range.shape, (2,1))

        Grad2D_chan = GradientOperator(self.ig_2D_chan, correlation="SpaceChannels", backend='c')
        numpy.testing.assert_equal(Grad2D_chan.range.shape, (3,1))

        Grad3D = GradientOperator(self.ig_3D, backend='c')
        numpy.testing.assert_equal(Grad3D.range.shape, (3,1))

        Grad3D_chan = GradientOperator(self.ig_3D_chan, correlation="SpaceChannels", backend='c')
        numpy.testing.assert_equal(Grad3D_chan.range.shape, (4,1))


    def test_GradientOperator_split_shape(self):
        for geom in [self.ig_2D, self.ig_2D_chan, self.ig_3D, self.ig_3D_chan]:

            for split in [True, False]:

                Grad = GradientOperator(geom, split = split, correlation="SpaceChannels", backend='c')

                if geom == self.ig_2D:
                    shape = (2,1)
                elif geom == self.ig_3D:
                    shape = (3,1)
                elif geom == self.ig_2D_chan:
                    if split:
                        shape = (2,1)
                    else:
                        shape = (3,1)
                elif geom == self.ig_3D_chan:
                    if split:
                        shape = (2,1)
                    else:
                        shape = (4,1)

                try:
                    numpy.testing.assert_equal(Grad.range.shape, shape)
                except:
                    self.print_assertion_info(geom, None, None, None, split=split)
                    raise


    def test_GradientOperator_split_direct_adjoint(self):
        # Test split for direct and adjoint in 2D + channels geometry
        geom = self.ig_2D_chan
        Grad2D_split_false = GradientOperator(geom, split = False, correlation="SpaceChannels", backend='c')
        Grad2D_split_true = GradientOperator(geom, split = True, correlation="SpaceChannels", backend='c')

        tmp_x = geom.allocate('random')
        res1 = Grad2D_split_false.direct(tmp_x)
        res2 = Grad2D_split_true.direct(tmp_x)
        res1_adj = Grad2D_split_false.adjoint(res1)
        res2_adj = Grad2D_split_true.adjoint(res2)

        numpy.testing.assert_array_almost_equal(res1[0].array, res2[0].array)
        numpy.testing.assert_array_almost_equal(res1[1].array, res2[1][0].array)
        numpy.testing.assert_array_almost_equal(res1[2].array, res2[1][1].array)
        numpy.testing.assert_array_almost_equal(res1_adj.array, res2_adj.array)

        # Test split for direct and adjoint in 3D + channels geometry
        geom = self.ig_3D_chan
        Grad3D_split_false = GradientOperator(geom, split = False, correlation="SpaceChannels", backend='c')
        Grad3D_split_true = GradientOperator(geom, split = True, correlation="SpaceChannels", backend='c')

        tmp_x = geom.allocate('random')
        res1 = Grad3D_split_false.direct(tmp_x)
        res2 = Grad3D_split_true.direct(tmp_x)
        res1_adj = Grad3D_split_false.adjoint(res1)
        res2_adj = Grad3D_split_true.adjoint(res2)

        numpy.testing.assert_array_almost_equal(res1[0].array, res2[0].array)
        numpy.testing.assert_array_almost_equal(res1[1].array, res2[1][0].array)
        numpy.testing.assert_array_almost_equal(res1[2].array, res2[1][1].array)
        numpy.testing.assert_array_almost_equal(res1[3].array, res2[1][2].array)
        numpy.testing.assert_array_almost_equal(res1_adj.array, res2_adj.array)


    def test_Gradient_operator_numpy_vs_c(self):

        #only use configurations supported by c backend
        config = self.test_backend_configurations[1]

        for geom in self.list_geometries:
            for bnd in config.get('bconditions'):
                for corr in config.get('correlation'):
                    for method in config.get('method'):

                        Grad_c = GradientOperator(geom, bnd_cond = bnd, method=method, correlation= corr, backend = 'c')
                        Grad_numpy =  GradientOperator(geom, bnd_cond = bnd, method=method, correlation= corr, backend = 'numpy')

                        tmp_x = geom.allocate('random')
                        res1_c = Grad_c.direct(tmp_x)
                        res1_np = Grad_numpy.direct(tmp_x)

                        # Check direct of numpy vs c in place
                        for m in range(len(Grad_c.range.geometries)):
                            try:
                                numpy.testing.assert_array_almost_equal(res1_c[m].array,
                                                                res1_np[m].array)
                            except:
                                self.print_assertion_info(geom, bnd, None, None, corr, None)
                                raise

                        res1_c_out = Grad_c.range.allocate()
                        Grad_c.direct(tmp_x, out = res1_c_out)

                        res1_np_out = Grad_numpy.range.allocate()
                        Grad_numpy.direct(tmp_x, out = res1_np_out)

                        # Check direct of numpy vs c allocate
                        for m in range(len(Grad_c.range.geometries)):
                            try:
                                numpy.testing.assert_array_almost_equal(res1_c_out[m].array,
                                                                res1_np_out[m].array)
                            except:
                                self.print_assertion_info(geom, bnd, None, None, corr, None)
                                raise

                        tmp_x = Grad_c.range.allocate('random')
                        res1_c = Grad_c.adjoint(tmp_x)
                        res1_np = Grad_numpy.adjoint(tmp_x)

                        # Check adjoint of numpy vs c in place
                        try:
                            numpy.testing.assert_array_almost_equal(res1_c.array,
                                                            res1_np.array, decimal=5)
                        except:
                            self.print_assertion_info(geom, bnd, None, None, corr, None)
                            raise

                        res1_c_out = Grad_c.domain.allocate()
                        Grad_c.adjoint(tmp_x, out = res1_c_out)

                        res1_np_out = Grad_numpy.domain.allocate()
                        Grad_numpy.adjoint(tmp_x, out = res1_np_out)

                        # Check adjoint of numpy vs c allocate
                        try:
                            numpy.testing.assert_array_almost_equal(res1_c_out.array,
                                                            res1_np_out.array, decimal=5)
                        except:
                            self.print_assertion_info(geom, bnd, None, None, corr, None)
                            raise


    def test_GradientOperator_for_pseudo_2D_geometries(self):
            numpy.random.seed(1)
            # ImageGeometry shape (5,5,1)
            ig1 = ImageGeometry(voxel_num_x = 1, voxel_num_y = 5, voxel_num_z=5,
                                voxel_size_x = 0.4, voxel_size_y = 0.2, voxel_size_z=0.6)
            # ImageGeometry shape (1,5,5)
            ig2 = ImageGeometry(voxel_num_x = 5, voxel_num_y = 5, voxel_num_z=1,
                                voxel_size_x = 0.1, voxel_size_y = 0.2, voxel_size_z=0.4)
            # ImageGeometry shape (5,1,5)
            ig3 = ImageGeometry(voxel_num_x = 5, voxel_num_y = 1, voxel_num_z=5,
                                voxel_size_x = 0.6, voxel_size_y = 0.4, voxel_size_z=0.3)

            data1 = ig1.allocate('random')
            data2 = ig2.allocate('random')
            data3 = ig3.allocate('random')

            data = [data1, data2, data3]
            ig = [ig1, ig2, ig3]

            for i in range(3):

                ########################################
                ##### Test Gradient numpy backend  #####
                ########################################
                Grad_numpy = GradientOperator(ig[i], backend='numpy')
                res1 = Grad_numpy.direct(data[i])
                res2 = Grad_numpy.range_geometry().allocate()
                Grad_numpy.direct(data[i], out=res2)

                # test direct with and without out
                numpy.testing.assert_array_almost_equal(res1[0].as_array(), res2[0].as_array())
                numpy.testing.assert_array_almost_equal(res1[1].as_array(), res2[1].as_array())

                # test adjoint with and without out
                res3 = Grad_numpy.adjoint(res1)
                res4 = Grad_numpy.domain_geometry().allocate()
                Grad_numpy.adjoint(res2, out=res4)
                numpy.testing.assert_array_almost_equal(res3.as_array(), res4.as_array())

                # test dot_test
                for sd in [5, 10, 15]:
                    self.assertTrue(LinearOperator.dot_test(Grad_numpy, seed=sd))

                # test shape of output of direct
                self.assertEqual(res1[0].shape, ig[i].shape)
                self.assertEqual(res1.shape, (2,1))

                ########################################
                ##### Test Gradient c backend  #####
                ########################################
                Grad_c = GradientOperator(ig[i], backend='c')

                # test direct with and without out
                res5 = Grad_c.direct(data[i])
                res6 = Grad_c.range_geometry().allocate()*0.
                Grad_c.direct(data[i], out=res6)

                numpy.testing.assert_array_almost_equal(res5[0].as_array(), res6[0].as_array())
                numpy.testing.assert_array_almost_equal(res5[1].as_array(), res6[1].as_array())

                # test adjoint
                res7 = Grad_c.adjoint(res5)
                res8 = Grad_c.domain_geometry().allocate()*0.
                Grad_c.adjoint(res5, out=res8)
                numpy.testing.assert_array_almost_equal(res7.as_array(), res8.as_array())

                # test dot_test
                for sd in [5, 10, 15]:
                    self.assertTrue(LinearOperator.dot_test(Grad_c, seed = sd))

                # test direct numpy vs direct c backends (with and without out)
                numpy.testing.assert_array_almost_equal(res5[0].as_array(), res1[0].as_array())
                numpy.testing.assert_array_almost_equal(res6[1].as_array(), res2[1].as_array())


    def test_GradientOperator_complex_data(self):
        # make complex dtype
        self.ig_2D.dtype = numpy.complex64
        x = self.ig_2D.allocate('random')

        Grad = GradientOperator(domain_geometry=self.ig_2D, backend='numpy')

        res1 = Grad.direct(x)
        res2 = Grad.range.allocate()
        Grad.direct(x, out=res2)

        numpy.testing.assert_array_almost_equal(res1[0].as_array(), res2[0].as_array())
        numpy.testing.assert_array_almost_equal(res1[1].as_array(), res2[1].as_array())

        # check dot_test
        for sd in [5, 10, 15]:
            self.assertTrue(LinearOperator.dot_test(Grad, seed=sd))


    def test_GradientOperator_cpp_failure_direct(self):
        # Simulate the failure by setting the status to non-zero
            ig = ImageGeometry(voxel_num_x = 2, voxel_num_y = 3, voxel_num_z=4)
            data = ig.allocate('random')

            Grad = GradientOperator(ig, backend='c')
        
            # with the call to status = self.fd(args) returning a non-zero value
            Grad.operator.fd = Mock(return_value=-1)
            with self.assertRaises(RuntimeError):
                Grad.direct(data)


    def test_GradientOperator_cpp_failure_adjoint(self):
        # Simulate the failure by setting the status to non-zero
            ig = ImageGeometry(voxel_num_x = 2, voxel_num_y = 3, voxel_num_z=4)
            data = ig.allocate('random')

            Grad = GradientOperator(ig, backend='c')
            res_direct = Grad.direct(data)
        
            # with the call to status = self.fd(args) returning a non-zero value
            Grad.operator.fd = Mock(return_value=-1)
            with self.assertRaises(RuntimeError):
                Grad.adjoint(res_direct)



#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import numpy as np
from cil.optimisation.operators import LinearOperator
import cil

from scipy.ndimage import convolve, correlate

class BlurringOperator(LinearOperator):

    r'''BlurringOperator:  D: X -> X,  takes in a numpy array PSF representing
    a point spread function for blurring the image. The implementation is
    generic and naive simply using convolution.

        :param PSF: numpy array with point spread function of blur.
        :param geometry: ImageGeometry of ImageData to work on.

     '''

    def __init__(self, PSF, geometry):
        super(BlurringOperator, self).__init__(domain_geometry=geometry,
                                           range_geometry=geometry)
        if isinstance(PSF,np.ndarray):
            self.PSF = PSF
        else:
            raise TypeError('PSF must be a number array with same number of dimensions as geometry.')

 

    def direct(self,x,out=None):
        '''Returns D(x). The forward mapping consists of convolution of the
        image with the specified PSF. Here reflective boundary conditions
        are selected.'''

        if out is None:
            result = self.range_geometry().allocate(0)
            result.fill(convolve(x.as_array(),self.PSF, mode='reflect'))
            return result
        else:
            outarr = out.as_array()
            convolve(x.as_array(),self.PSF, output=outarr, mode='reflect')
            out.fill(outarr)
            return out

    def adjoint(self,x, out=None):

        '''Returns D^{*}(y). The adjoint of convolution is convolution with
        the PSF rotated by 180 degrees, or equivalently correlation by the PSF
        itself.'''

        if out is None:
            result = self.domain_geometry().allocate(0)
            result.fill(correlate(x.as_array(),self.PSF, mode='reflect'))
            return result
        else:
            outarr = out.as_array()
            correlate(x.as_array(),self.PSF, output=outarr, mode='reflect')
            out.fill(outarr)
            return out




#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import warnings

from numbers import Number
import numpy as np
from functools import reduce
from cil.utilities.errors import InPlaceError


class Function(object):

    r""" Abstract class representing a function

        Parameters
        ----------

        L: number, positive, default None
            Lipschitz constant of the gradient of the function F(x), when it is differentiable.

        Note
        -----
        The Lipschitz of the gradient of the function is a positive real number, such that :math:`\|f'(x) - f'(y)\| \leq L \|x-y\|`, assuming :math:`f: IG \rightarrow \mathbb{R}`

    """

    def __init__(self, L=None):
        # overrides the type check to allow None as initial value
        self._L = L

    def __call__(self, x):

        raise NotImplementedError

    def gradient(self, x, out=None):
        r"""Returns the value of the gradient of function :math:`F`  evaluated at :math:`x`, if it is differentiable

        .. math:: F'(x)

        Parameters
        ----------
        x : DataContainer

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        --------
        DataContainer, the value of the gradient of the function at x.

        """
        raise NotImplementedError

    def proximal(self, x, tau, out=None):
        r"""Returns the proximal operator of function :math:`\tau F`  evaluated at x

        .. math:: \text{prox}_{\tau F}(x) = \underset{z}{\text{argmin}} \frac{1}{2}\|z - x\|^{2} + \tau F(z)

        Parameters
        ----------
        x : DataContainer

        tau: scalar

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the proximal operator of the function at x with scalar :math:`\tau`.

        """
        raise NotImplementedError

    def convex_conjugate(self, x):
        r""" Evaluation of the function F* at x, where F* is the convex conjugate of function F,

        .. math:: F^{*}(x^{*}) = \underset{x}{\sup} \langle x^{*}, x \rangle - F(x)

        Parameters
        ----------
        x : DataContainer

        Returns
        -------
        The value of the convex conjugate of the function at x.

        """
        raise NotImplementedError

    def proximal_conjugate(self, x, tau, out=None):
        r"""Returns the proximal operator of the convex conjugate of function :math:`\tau F` evaluated at :math:`x^{*}`

        .. math:: \text{prox}_{\tau F^{*}}(x^{*}) = \underset{z^{*}}{\text{argmin}} \frac{1}{2}\|z^{*} - x^{*}\|^{2} + \tau F^{*}(z^{*})

        Due to Moreau’s identity, we have an analytic formula to compute the proximal operator of the convex conjugate :math:`F^{*}`

        .. math:: \text{prox}_{\tau F^{*}}(x) = x - \tau\text{prox}_{\tau^{-1} F}(\tau^{-1}x)

        Parameters
        ----------
        x : DataContainer

        tau: scalar

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the value of the proximal operator of the convex conjugate at point :math:`x` for scalar :math:`\tau` or None if `out`.

        """
        if id(x) == id(out):
            raise InPlaceError(
                message="The proximal_conjugate of a CIL function cannot be used in place")

        try:
            tmp = x
            x.divide(tau, out=tmp)
        except TypeError:
            tmp = x.divide(tau, dtype=np.float32)

        val = self.proximal(tmp, 1.0/tau, out=out)

        if id(tmp) == id(x):
            x.multiply(tau, out=x)

        val.sapyb(-tau,  x, 1.0, out=val)

        return val

    # Algebra for Function Class

        # Add functions
        # Subtract functions
        # Add/Substract with Scalar
        # Multiply with Scalar

    def __add__(self, other):
        """ Returns the sum of the functions.

            Cases: a) the sum of two functions :math:`(F_{1}+F_{2})(x) = F_{1}(x) + F_{2}(x)`
                   b) the sum of a function with a scalar :math:`(F_{1}+scalar)(x) = F_{1}(x) + scalar`

        """

        if isinstance(other,  Number):
            return SumScalarFunction(self, other)
        return SumFunction(self, other)

    def __radd__(self, other):
        """ Making addition commutative. """
        return self + other

    def __sub__(self, other):
        """ Returns the subtraction of the functions."""
        return self + (-1) * other

    def __rmul__(self, scalar):
        """Returns a function multiplied by a scalar."""
        return ScaledFunction(self, scalar)

    def __mul__(self, scalar):
        return self.__rmul__(scalar)

    def __neg__(self):
        """ Return the negative of the function """
        return -1 * self


    def centered_at(self, center):
        """ Returns a translated function, namely if we have a function :math:`F(x)` the center is at the origin.
            TranslateFunction is :math:`F(x - b)` and the center is at point b.

        Parameters
        ----------
        center: DataContainer
            The point to center the function at.

        Returns
        -------
        The translated function.
        """

        if center is None:
            return self
        else:
            return TranslateFunction(self, center)

    @property
    def L(self):
        r'''Lipschitz of the gradient of function f.

        L is positive real number, such that :math:`\|f'(x) - f'(y)\| \leq L\|x-y\|`, assuming :math:`f: IG \rightarrow \mathbb{R}`'''
        return self._L
        # return self._L

    @L.setter
    def L(self, value):
        '''Setter for Lipschitz constant'''
        if isinstance(value, (Number,)) and value >= 0:
            self._L = value
        else:
            raise TypeError('The Lipschitz constant is a real positive number')


class SumFunction(Function):

    r"""SumFunction represents the sum of :math:`n\geq2` functions

    .. math:: (F_{1} + F_{2} + ... + F_{n})(\cdot)  = F_{1}(\cdot) + F_{2}(\cdot) + ... + F_{n}(\cdot)

    Parameters
    ----------

    *functions : Functions
                 Functions to set up a :class:`.SumFunction`

    Raises
    ------
    ValueError
            If the number of function is strictly less than 2.


    Examples
    --------
    .. math:: F(x) = \|x\|^{2} + \frac{1}{2}\|x - 1\|^{2}

    >>> from cil.optimisation.functions import L2NormSquared
    >>> from cil.framework import ImageGeometry
    >>> f1 = L2NormSquared()
    >>> f2 = 0.5 * L2NormSquared(b = ig.allocate(1))
    >>> F = SumFunction(f1, f2)

    .. math:: F(x) = \sum_{i=1}^{50} \|x - i\|^{2}

    >>> F = SumFunction(*[L2NormSquared(b=i) for i in range(50)])


    """

    def __init__(self, *functions):

        super(SumFunction, self).__init__()
        if not len(functions):
            raise IndexError('At least 1 function needed')
        self.functions = functions

    @property
    def L(self):
        """Returns the Lipschitz constant for the SumFunction

        .. math:: L = \sum_{i} L_{i}

        where :math:`L_{i}` is the Lipschitz constant of the smooth function :math:`F_{i}`.

        """

        L = 0.
        for f in self.functions:
            if f.L is not None:
                L += f.L
            else:
                L = None
                break
        self._L = L

        return self._L

    @L.setter
    def L(self, value):
        # call base class setter
        super(SumFunction, self.__class__).L.fset(self, value)

    @property
    def Lmax(self):
        """Returns the maximum Lipschitz constant for the SumFunction

        .. math:: L = \max_{i}\{L_{i}\}

        where :math:`L_{i}` is the Lipschitz constant of the smooth function :math:`F_{i}`.

        """

        l = []
        for f in self.functions:
            if f.L is not None:
                l.append(f.L)
            else:
                l = None
                break
        self._Lmax = max(l)

        return self._Lmax

    @Lmax.setter
    def Lmax(self, value):
        # call base class setter
        super(SumFunction, self.__class__).Lmax.fset(self, value)

    def __call__(self, x):
        r"""Returns the value of the sum of functions evaluated at :math:`x`.

        .. math:: (F_{1} + F_{2} + ... + F_{n})(x) = F_{1}(x) + F_{2}(x) + ... + F_{n}(x)

        """
        ret = 0.
        for f in self.functions:
            ret += f(x)
        return ret

    def gradient(self, x, out=None):
        r"""Returns the value of the sum of the gradient of functions evaluated at :math:`x`, if all of them are differentiable.

        .. math:: (F'_{1} + F'_{2} + ... + F'_{n})(x) = F'_{1}(x) + F'_{2}(x) + ... + F'_{n}(x)

        Parameters
        ----------
        x : DataContainer
            Point to evaluate the gradient at.
        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the value of the sum of the gradients evaluated at point :math:`x`.

        """
        if out is not None and id(x) == id(out):
            raise InPlaceError

        for i, f in enumerate(self.functions):
            if i == 0:
                ret = f.gradient(x, out=out)
            else:
                ret += f.gradient(x)
        return ret

    def __add__(self, other):
        """ Addition for the SumFunction.

        *  :code:`SumFunction` + :code:`SumFunction` is a :code:`SumFunction`.

        *  :code:`SumFunction` + :code:`Function` is a :code:`SumFunction`.

        """

        if isinstance(other, SumFunction):
            functions = list(self.functions) + list(other.functions)
            return SumFunction(*functions)
        elif isinstance(other, Function):
            functions = list(self.functions)
            functions.append(other)
            return SumFunction(*functions)
        else:
            return super(SumFunction, self).__add__(other)

    @property
    def num_functions(self):
        return len(self.functions)


class ScaledFunction(Function):

    r""" ScaledFunction represents the scalar multiplication with a Function.

    Let a function F then and a scalar :math:`\alpha`.

    If :math:`G(x) = \alpha F(x)` then:

    1. :math:`G(x) = \alpha  F(x)` ( __call__ method )
    2. :math:`G'(x) = \alpha  F'(x)` ( gradient method )
    3. :math:`G^{*}(x^{*}) = \alpha  F^{*}(\frac{x^{*}}{\alpha})` ( convex_conjugate method )
    4. :math:`\text{prox}_{\tau G}(x) = \text{prox}_{(\tau\alpha) F}(x)` ( proximal method )

    """

    def __init__(self, function, scalar):

        super(ScaledFunction, self).__init__()

        if not isinstance(scalar, Number):
            raise TypeError('expected scalar: got {}'.format(type(scalar)))

        self.scalar = scalar
        self.function = function

    @property
    def L(self):
        if self._L is None:
            if self.function.L is not None:
                self._L = abs(self.scalar) * self.function.L
            else:
                self._L = None
        return self._L

    @L.setter
    def L(self, value):
        # call base class setter
        super(ScaledFunction, self.__class__).L.fset(self, value)

    @property
    def scalar(self):
        return self._scalar

    @scalar.setter
    def scalar(self, value):
        if isinstance(value, (Number, )):
            self._scalar = value
        else:
            raise TypeError(
                'Expecting scalar type as a number type. Got {}'.format(type(value)))

    def __call__(self, x):
        r"""Returns the value of the scaled function evaluated at :math:`x`.

        .. math:: G(x) = \alpha F(x)

        Parameters
        ----------
        x : DataContainer

        Returns
        --------
        DataContainer, the value of the scaled function.
        """
        return self.scalar * self.function(x)

    def convex_conjugate(self, x):
        r"""Returns the convex conjugate of the scaled function.

        .. math:: G^{*}(x^{*}) = \alpha  F^{*}(\frac{x^{*}}{\alpha})

        Parameters
        ----------
        x : DataContainer

        Returns
        -------
        The value of the convex conjugate of the scaled function.

        """
        try:
            x.divide(self.scalar, out=x)
            tmp = x
        except TypeError:
            tmp = x.divide(self.scalar, dtype=np.float32)

        val = self.function.convex_conjugate(tmp)

        if id(tmp) == id(x):
            x.multiply(self.scalar, out=x)

        return self.scalar * val

    def gradient(self, x, out=None):
        r"""Returns the gradient of the scaled function evaluated at :math:`x`.

        .. math:: G'(x) = \alpha  F'(x)

        Parameters
        ----------
        x : DataContainer
            Point to evaluate the gradient at.
        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the value of the gradient of the scaled function evaluated at :math:`x`.

        """
        res = self.function.gradient(x, out=out)
        res *= self.scalar
        return res

    def proximal(self, x, tau, out=None):
        r"""Returns the proximal operator of the scaled function, evaluated at :math:`x`.

        .. math:: \text{prox}_{\tau G}(x) = \text{prox}_{(\tau\alpha) F}(x)

        Parameters
        ----------
        x : DataContainer

        tau: scalar

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the proximal operator of the scaled function evaluated at :math:`x` with scalar :math:`\tau`.

        """

        return self.function.proximal(x, tau*self.scalar, out=out)

    def proximal_conjugate(self, x, tau, out=None):
        r"""This returns the proximal  conjugate operator for the function at :math:`x`, :math:`\tau`

        Parameters
        ----------
        x : DataContainer

        tau: scalar

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the proximal conjugate operator for the function evaluated at :math:`x` and :math:`\tau`.

        """
        if out is not None and id(x) == id(out):
            raise InPlaceError

        try:
            tmp = x
            x.divide(tau, out=tmp)
        except TypeError:
            tmp = x.divide(tau, dtype=np.float32)

        val = self.function.proximal(tmp, self.scalar/tau, out=out)

        if id(tmp) == id(x):
            x.multiply(tau, out=x)

        val.sapyb(-tau,  x, 1.0, out=val)

        return val


class SumScalarFunction(SumFunction):

    """ SumScalarFunction represents the sum a function with a scalar.

        .. math:: (F + scalar)(x)  = F(x) + scalar

        Although SumFunction has no general expressions for

        i) convex_conjugate
        ii) proximal
        iii) proximal_conjugate

        if the second argument is a ConstantFunction then we can derive the above analytically.

    """

    def __init__(self, function, constant):

        super(SumScalarFunction, self).__init__(
            function, ConstantFunction(constant))
        self.constant = constant
        self.function = function

    def convex_conjugate(self, x):
        r""" Returns the convex conjugate of a :math:`(F+scalar)`, evaluated at :math:`x`.

        .. math:: (F+scalar)^{*}(x^{*}) = F^{*}(x^{*}) - scalar

        Parameters
        ----------
        x : DataContainer

        Returns
        -------
        The value of the convex conjugate evaluated at :math:`x`.

        """
        return self.function.convex_conjugate(x) - self.constant

    def proximal(self, x, tau, out=None):
        r""" Returns the proximal operator of :math:`F+scalar`

        .. math:: \text{prox}_{\tau (F+scalar)}(x) = \text{prox}_{\tau F}

        Parameters
        ----------
        x : DataContainer

        tau: scalar

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the evaluation of the proximal operator evaluated at :math:`x` and :math:`\tau`.

        """
        return self.function.proximal(x, tau, out=out)

    @property
    def L(self):
        if self._L is None:
            if self.function.L is not None:
                self._L = self.function.L
            else:
                self._L = None
        return self._L

    @L.setter
    def L(self, value):
        # call base class setter
        super(SumScalarFunction, self.__class__).L.fset(self, value)


class ConstantFunction(Function):

    r""" ConstantFunction: :math:`F(x) = constant, constant\in\mathbb{R}`

    """

    def __init__(self, constant=0):
        self.constant = constant
        super(ConstantFunction, self).__init__(L=1)

    def __call__(self, x):
        """ Returns the value of the function, :math:`F(x) = constant`"""
        return self.constant

    def gradient(self, x, out=None):
        """ Returns the value of the gradient of the function, :math:`F'(x)=0`
        Parameters
        ----------
        x : DataContainer
            Point to evaluate the gradient at.
        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        A DataContainer of zeros, the same size as :math:`x`.

        """
        if out is None:
            return x * 0.
        else:
            out.fill(0)
            return out

    def convex_conjugate(self, x):
        r""" The convex conjugate of constant function :math:`F(x) = c\in\mathbb{R}` is

        .. math::
            F(x^{*})
            =
            \begin{cases}
                -c, & if x^{*} = 0\\
                \infty, & \mbox{otherwise}
            \end{cases}


        However, :math:`x^{*} = 0` only in the limit of iterations, so in fact this can be infinity.
        We do not want to have inf values in the convex conjugate, so we have to penalise this value accordingly.
        The following penalisation is useful in the PDHG algorithm, when we compute primal & dual objectives
        for convergence purposes.

        .. math:: F^{*}(x^{*}) = \sum \max\{x^{*}, 0\}

        Parameters
        ----------
        x : DataContainer

        Returns
        -------
        The maximum of x and 0, summed over the entries of x.

        """
        return x.maximum(0).sum()

    def proximal(self, x, tau, out=None):
        r"""Returns the proximal operator of the constant function, which is the same element, i.e.,

        .. math:: \text{prox}_{\tau F}(x) = x

        Parameters
        ----------
        x : DataContainer

        tau: scalar

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, equal to :math:`x`.

        """
        if out is None:
            return x.copy()
        else:
            out.fill(x)
            return out

    @property
    def constant(self):
        return self._constant

    @constant.setter
    def constant(self, value):
        if not isinstance(value, Number):
            raise TypeError('expected scalar: got {}'.format(type(value)))
        self._constant = value

    @property
    def L(self):
        return 1.

    def __rmul__(self, other):
        '''defines the right multiplication with a number'''
        if not isinstance(other, Number):
            raise NotImplemented
        constant = self.constant * other
        return ConstantFunction(constant)


class ZeroFunction(ConstantFunction):

    """ ZeroFunction represents the zero function, :math:`F(x) = 0`
    """

    def __init__(self):
        super(ZeroFunction, self).__init__(constant=0.)


class TranslateFunction(Function):

    r""" TranslateFunction represents the translation of function F with respect to the center b.

    Let a function F and consider :math:`G(x) = F(x - center)`.

    Function F is centered at 0, whereas G is centered at point b.

    If :math:`G(x) = F(x - b)` then:

    1. :math:`G(x) = F(x - b)` ( __call__ method )
    2. :math:`G'(x) = F'(x - b)` ( gradient method )
    3. :math:`G^{*}(x^{*}) = F^{*}(x^{*}) + <x^{*}, b >` ( convex_conjugate method )
    4. :math:`\text{prox}_{\tau G}(x) = \text{prox}_{\tau F}(x - b)  + b` ( proximal method )

    """

    def __init__(self, function, center):
        try:
            L = function.L
        except NotImplementedError as nie:
            L = None
        super(TranslateFunction, self).__init__(L=L)

        self.function = function
        self.center = center

    def __call__(self, x):
        r"""Returns the value of the translated function.

        .. math:: G(x) = F(x - b)

        Parameters
        ----------
        x : DataContainer

        Returns
        -------
        The value of the translated function evaluated at :math:`x`.


        """
        try:
            x.subtract(self.center, out=x)
            tmp = x
        except TypeError:
            tmp = x.subtract(self.center, dtype=np.float32)

        val = self.function(tmp)

        if id(tmp) == id(x):
            x.add(self.center, out=x)

        return val

    def gradient(self, x, out=None):
        r"""Returns the gradient of the translated function.

        .. math:: G'(x) =  F'(x - b)

        Parameters
        ----------
        x : DataContainer
            Point to evaluate the gradient at.

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the gradient of the translated function evaluated at :math:`x`.
        """

        if id(x) == id(out):
            raise InPlaceError

        try:
            x.subtract(self.center, out=x)
            tmp = x
        except TypeError:
            tmp = x.subtract(self.center, dtype=np.float32)

        val = self.function.gradient(tmp, out=out)

        if id(tmp) == id(x):
            x.add(self.center, out=x)

        return val

    def proximal(self, x, tau, out=None):
        r"""Returns the proximal operator of the translated function.

        .. math:: \text{prox}_{\tau G}(x) = \text{prox}_{\tau F}(x-b) + b

        Parameters
        ----------
        x : DataContainer

        tau: scalar

        out: return DataContainer, if None a new DataContainer is returned, default None.

        Returns
        -------
        DataContainer, the proximal operator of the translated function at :math:`x` and :math:`\tau`.
        """

        if id(x) == id(out):
            raise InPlaceError

        try:
            x.subtract(self.center, out=x)
            tmp = x
        except TypeError:
            tmp = x.subtract(self.center, dtype=np.float32)

        val = self.function.proximal(tmp, tau, out=out)
        val.add(self.center, out=val)

        if id(tmp) == id(x):
            x.add(self.center, out=x)

        return val

    def convex_conjugate(self, x):
        r"""Returns the convex conjugate of the translated function.

        .. math:: G^{*}(x^{*}) = F^{*}(x^{*}) + <x^{*}, b >

        Parameters
        ----------
        x : DataContainer

        Returns
        -------
        The value of the convex conjugate of the translated function at :math:`x`.

        """

        return self.function.convex_conjugate(x) + self.center.dot(x)



#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


from cil.optimisation.algorithms import Algorithm
from cil.optimisation.functions import ZeroFunction
import logging
import warnings
class PD3O(Algorithm):
    

    r"""Primal Dual Three Operator Splitting (PD3O) algorithm, see "A New Primal–Dual Algorithm for Minimizing the Sum
        of Three Functions with a Linear Operator".  This is a primal dual algorithm for minimising :math:`f(x)+g(x)+h(Ax)` where all functions are proper, lower semi-continuous and convex, 
        :math:`f` should be differentiable with a Lipschitz continuous gradient and :math:`A` is a bounded linear operator. 
    
        Parameters
        ----------
        f : Function
            A smooth function with Lipschitz continuous gradient.
        g : Function
            A convex function with a computationally computable proximal.
        h : Function
            A composite convex function.
        operator: Operator
            Bounded linear operator
        delta: Float, optional, default is  `1./(gamma*operator.norm()**2)`
            The dual step-size 
        gamma: Float, optional, default is `2.0/f.L`
            The primal step size 
        initial : DataContainer, optional default is a container of zeros, in the domain of the operator 
            Initial point for the  algorithm.             


        Reference
        ---------
        Yan, M. A New Primal–Dual Algorithm for Minimizing the Sum of Three Functions with a Linear Operator. J Sci Comput 76, 1698–1717 (2018). https://doi.org/10.1007/s10915-018-0680-3
     """    


    def __init__(self, f, g, h, operator, delta=None, gamma=None, initial=None, **kwargs):

        super(PD3O, self).__init__(**kwargs)

              
        self.set_up(f=f, g=g, h=h,  operator=operator, delta=delta, gamma=gamma, initial=initial, **kwargs)
 
                  
    def set_up(self, f, g, h, operator, delta=None, gamma=None, initial=None,**kwargs):
        
        logging.info("{} setting up".format(self.__class__.__name__, ))
        
        self.f = f # smooth function
        if isinstance(self.f, ZeroFunction):
            warnings.warn(" If self.f is the ZeroFunction, then PD3O = PDHG. Please use PDHG instead. Otherwise, select a relatively small parameter gamma ", UserWarning)
            if gamma is None:
                gamma = 1.0/operator.norm()                      
        
        self.g = g # proximable
        self.h = h # composite
        self.operator = operator
        
        if gamma is None:
            gamma = 0.99*2.0/self.f.L
        
        if delta is None :
            delta = 0.99/(gamma*self.operator.norm()**2)
        
        self.gamma = gamma
        self.delta = delta  

        if initial is None:
            self.x = self.operator.domain_geometry().allocate(0)
        else:
            self.x = initial.copy()

        self.x_old = self.x.copy()    
        
        self.s_old = self.operator.range_geometry().allocate(0)
        self.s = self.operator.range_geometry().allocate(0)
                
        self.grad_f = self.operator.domain_geometry().allocate(0)        
  
        self.configured = True
        logging.info("{} configured".format(self.__class__.__name__, ))
        
        # initial proximal conjugate step
        self.operator.direct(self.x_old, out=self.s)
        self.s_old.sapyb(1, self.s, self.delta, out=self.s_old)
        self.h.proximal_conjugate(self.s_old, self.delta, out=self.s)
        

    def update(self):
        r""" Performs a single iteration of the PD3O algorithm        
        """

        # Following equations 4 in https://link.springer.com/article/10.1007/s10915-018-0680-3
        # in this case order of proximal steps we recover the (primal) PDHG, when f=0

        
        tmp = self.x_old
        self.x_old = self.x
        self.x = tmp
        
        
        # proximal step        
        self.f.gradient(self.x_old, out=self.grad_f)
        self.x_old.sapyb(1., self.grad_f, -self.gamma, out = self.grad_f) # x_old - gamma * grad_f(x_old)        
        self.operator.adjoint(self.s, out=self.x_old)
        self.x_old.sapyb(-self.gamma, self.grad_f, 1.0, out=self.x_old)
        self.g.proximal(self.x_old, self.gamma, out = self.x)
    
        # update step        
        self.f.gradient(self.x, out=self.x_old)                    
        self.x_old *= self.gamma
        self.grad_f += self.x_old
        self.x.sapyb(2, self.grad_f, -1.0,  out=self.x_old) # 2*x - x_old + gamma*(grad_f_x_old) - gamma*(grad_f_x)
        
        tmp = self.s_old
        self.s_old = self.s
        self.s = tmp
        
        # proximal conjugate step
        self.operator.direct(self.x_old, out=self.s)
        self.s_old.sapyb(1, self.s, self.delta, out=self.s_old)
        self.h.proximal_conjugate(self.s_old, self.delta, out=self.s)
        
        
               
           
                                                                        
    def update_objective(self):
        """
        Evaluates the primal objective
        """
        self.operator.direct(self.x, out=self.s_old)        
        fun_h = self.h(self.s_old)         
        fun_g = self.g(self.x)
        fun_f = self.f(self.x)
        p1 = fun_f + fun_g + fun_h
        
        self.loss.append(p1)
        
        
        


#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# - CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# - Daniel Deidda (National Physical Laboratory, UK)
# - Claire Delplancke (Electricite de France, Research and Development)
# - Ashley Gillman (Australian e-Health Res. Ctr., CSIRO, Brisbane, Queensland, Australia)
# - Zeljko Kereta (Department of Computer Science, University College London, UK)
# - Evgueni Ovtchinnikov (STFC - UKRI)
# - Georg Schramm (Department of Imaging and Pathology, Division of Nuclear Medicine, KU Leuven, Leuven, Belgium)



from cil.optimisation.functions import SumFunction
from cil.optimisation.utilities import Sampler
import numbers
from abc import ABC, abstractmethod
import numpy as np


class ApproximateGradientSumFunction(SumFunction, ABC):
    r"""ApproximateGradientSumFunction represents the following sum 

    .. math:: \sum_{i=0}^{n-1} f_{i} = (f_{0} + f_{2} + ... + f_{n-1})

    where there are :math:`n` functions. This function class has two ways of calling gradient
    
        - `full_gradient` calculates the gradient of the sum :math:`\sum_{i=0}^{n-1} \nabla f_{i}`
        - `gradient` calls an `approximate_gradient` function which may be less computationally expensive to calculate than the full gradient
    
    

    This class is an abstract class.

    Parameters
    -----------
    functions : `list`  of functions
        A list of functions: :math:`[f_{0}, f_{2}, ..., f_{n-1}]`. Each function is assumed to be smooth with an implemented :func:`~Function.gradient` method. All functions must have the same domain. The number of functions (equivalently the length of the list) must be strictly greater than 1. 
    sampler: An instance of a CIL Sampler class ( :meth:`~optimisation.utilities.sampler`) or of another class which has a :code:`__next__` function implemented to output integers in :math:`{0,...,n-1}`.
        This sampler is called each time :code:`gradient` is called and sets the internal :code:`function_num` passed to the :code:`approximate_gradient` function.  Default is :code:`Sampler.random_with_replacement(len(functions))`. 

    Note
    -----
    We ensure that the approximate gradient is of a similar order of magnitude to the full gradient calculation. For example, in the :code:`SGFunction` we approximate the full gradient by :math:`n\nabla f_i` for an index :math:`i` given by the sampler. 
    The multiplication by :math:`n` is a choice to more easily allow comparisons between stochastic and non-stochastic methods and between stochastic methods with varying numbers of subsets. 

    Note
    -----
    Each time :code:`gradient` is called the class keeps track of which functions have been used to calculate the gradient. This may be useful for debugging or plotting after using this function in an iterative algorithm.  
    
        - The property :code:`data_passes_indices` is a list of lists holding the indices of the functions that are processed in each call of `gradient`. This list is updated each time `gradient` is called by appending a list of the indices of the functions used to calculate the gradient. 
        - The property :code:`data_passes` is a list of floats that holds the amount of data that has been processed up until each call of `gradient`. This list is updated each time `gradient` is called by appending the proportion of the data used when calculating the approximate gradient since the class was initialised (a full gradient calculation would be 1 full data pass). Warning: if your functions do not contain an equal `amount` of data, for example your data was not partitioned into equal batches, then you must first use the `set_data_partition_weights" function for this to be accurate.    



    Note
    ----
    The :meth:`~ApproximateGradientSumFunction.gradient` returns the approximate gradient depending on an index provided by the  :code:`sampler` method. 

    Example
    -------
    This class is an abstract base class, so we give an example using the SGFunction child class. 
    
    Consider the objective is to minimise: 

    .. math:: \sum_{i=0}^{n-1} f_{i}(x) = \sum_{i=0}^{n-1}\|A_{i} x - b_{i}\|^{2}

    >>> list_of_functions = [LeastSquares(Ai, b=bi)] for Ai,bi in zip(A_subsets, b_subsets))   
    >>> f = ApproximateGradientSumFunction(list_of_functions) 

    >>> list_of_functions = [LeastSquares(Ai, b=bi)] for Ai,bi in zip(A_subsets, b_subsets)) 
    >>> sampler = Sampler.sequential(len(list_of_functions))
    >>> f = SGFunction(list_of_functions, sampler=sampler)   
    >>> f.full_gradient(x)
    This will return :math:`\sum_{i=0}^{n-1} \nabla f_{i}(x)`
    >>> f.gradient(x)
    As per the approximate gradient implementation in the SGFunction this will return :math:`\nabla f_{0}`. The choice of the `0` index is because we chose a `sequential` sampler and this is the first time we called `gradient`. 
    >>> f.gradient(x)
    This will return :math:`\nabla f_{1}` because we chose  a `sequential` sampler and this is the second time we called `gradient`. 



    """

    def __init__(self, functions, sampler=None):

        if sampler is None:
            sampler = Sampler.random_with_replacement(len(functions))

        if not isinstance(functions, list):
            raise TypeError("Input to functions should be a list of functions")
        if not hasattr(sampler, "next"):
            raise ValueError('The provided sampler must have a `next` method')

        self.sampler = sampler

        self._partition_weights = [1 / len(functions)] * len(functions)

        self._data_passes_indices = []

        super(ApproximateGradientSumFunction, self).__init__(*functions)

    def __call__(self, x):
        r"""Returns the value of the sum of functions at :math:`x`.

        .. math:: (f_{0} + f_{1} + ... + f_{n-1})(x) = f_{0}(x) + f_{1}(x) + ... + f_{n-1}(x)

        Parameters
        ----------
        x : DataContainer

        --------
        float
            the value of the SumFunction at x


        """
        return super(ApproximateGradientSumFunction, self).__call__(x)

    def full_gradient(self, x, out=None):
        r"""Returns the value of the  full gradient of the sum of functions at :math:`x`.

        .. math:: \nabla_x(f_{0} + f_{1} + ... + f_{n-1})(x) = \nabla_xf_{0}(x) + \nabla_xf_{1}(x) + ... + \nabla_xf_{n-1}(x)

        Parameters
        ----------
        x : DataContainer
        out: return DataContainer, if `None` a new DataContainer is returned, default `None`.

        Returns
        --------
        DataContainer
            The value of the gradient of the sum function at x or nothing if `out`  
        """

        return super(ApproximateGradientSumFunction, self).gradient(x, out=out)

    @abstractmethod
    def approximate_gradient(self, x, function_num,   out=None):
        """ Returns the approximate gradient at a given point :code:`x` given a `function_number` in {0,...,len(functions)-1}.

        Parameters
        ----------
        x : DataContainer
        out: return DataContainer, if `None` a new DataContainer is returned, default `None`.
        function_num: `int` 
            Between 0 and the number of functions in the list  
        Returns
        --------
        DataContainer
            the value of the approximate gradient of the sum function at :code:`x` given a `function_number` in {0,...,len(functions)-1} 
        """
        pass

    def gradient(self, x, out=None):
        """ Selects a random function using the `sampler` and then calls the approximate gradient at :code:`x`

        Parameters
        ----------
        x : DataContainer
        out: return DataContainer, if `None` a new DataContainer is returned, default `None`.

        Returns
        --------
        DataContainer
            the value of the approximate gradient of the sum function at :code:`x`   
        """

        self.function_num = self.sampler.next()
        
        self._update_data_passes_indices([self.function_num])
        

        
        return self.approximate_gradient(x, self.function_num, out=out)
        

    def _update_data_passes_indices(self, indices):
        """ Internal function that updates the list of lists containing the function indices used to calculate the approximate gradient. 
        Parameters
        ----------
        indices: list
            List of indices used to calculate the approximate gradient in a given iteration

        """
        self._data_passes_indices.append(indices)

    def set_data_partition_weights(self, weights):
        """ Setter for the partition weights used to calculate the data passes  

        Parameters
        ----------
        weights: list of positive floats that sum to one. 
            The proportion of the data held in each function. Equivalent to the proportions that you partitioned your data into. 

        """
        if len(weights) != len(self.functions):
            raise ValueError(
                'The provided weights must be a list the same length as the number of functions')

        if abs(sum(weights) - 1) > 1e-6:
            raise ValueError('The provided weights must sum to one')

        if any(np.array(weights) < 0):
            raise ValueError(
                'The provided weights must be greater than or equal to zero')

        self._partition_weights = weights

    @property
    def data_passes_indices(self):
        """ The property :code:`data_passes_indices` is a list of lists holding the indices of the functions that are processed in each call of `gradient`. This list is updated each time `gradient` is called by appending a list of the indices of the functions used to calculate the gradient.  """
        return self._data_passes_indices

    @property
    def data_passes(self):
        """ The property :code:`data_passes` is a list of floats that holds the amount of data that has been processed up until each call of `gradient`. 
        This list is updated each time `gradient` is called by appending the proportion of the data used when calculating the approximate gradient since the class was initialised (a full gradient calculation would be 1 full data pass). 
        Note that if your functions do not contain an equal `amount` of data, for example your data was not partitioned into equal batches, then you must first use the `set_data_partition_weights` function for this to be accurate.   """
        data_passes = []
        for el in self.data_passes_indices:
            try:
                data_passes.append(data_passes[-1])
            except IndexError:
                data_passes.append(0)
            for i in el:
                data_passes[-1] += self._partition_weights[i]
        return data_passes
